#define WIN32_LEAN_AND_MEAN
#define COBJMACROS
#include <windows.h>
#include <shlobj.h>
#include "shellbrowser.h"
#include "commctrl.h"
#include "resid.h"
#include "debug.h"

static const WCHAR BROWSEWND_CLASS_NAME[] = {'_','S','h','e','l','l','W','i','n','d','o','w',0};

typedef struct tagShellBrowser
{
	IShellBrowserVtbl *lpVtbl;
	IOleCommandTargetVtbl *lpIOleCommandTargetVtbl;
	ULONG refcount;
	HWND hwndSelf;
	HWND hwndView;
	HWND hwndStatus;
	IShellFolder *folder;
	IShellView *view, *activeView;
} ShellBrowser;

extern IShellBrowserVtbl ShellBrowserVtbl;
extern IOleCommandTargetVtbl ShellBrowser_IOleCommandTargetVtbl;

static inline ShellBrowser *impl_from_IOleCommandTarget(IOleCommandTarget *iface)
{
    return (ShellBrowser *)((char *)iface - FIELD_OFFSET(ShellBrowser, lpIOleCommandTargetVtbl));
}

static inline IOleCommandTarget *_IOleCommandTarget_(ShellBrowser *This)
{
	return (IOleCommandTarget *)(&This->lpIOleCommandTargetVtbl);
}

WNDPROC old_status_proc;

LRESULT WINAPI status_proc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	WCHAR *ptr = (char *)lParam;

	if (uMsg == WM_PAINT || uMsg == WM_ERASEBKGND || uMsg == WM_NCPAINT)
		return CallWindowProc(old_status_proc, hWnd, uMsg, wParam, lParam);

	return CallWindowProc(old_status_proc, hWnd, uMsg, wParam, lParam);
}

ShellBrowser *ShellBrowser_Create(HWND hwnd)
{
	ShellBrowser *ret = CoTaskMemAlloc(sizeof(ShellBrowser));
	FOLDERSETTINGS settings;
	const INT status_widths[3] = {-1, 75, 140};
	RECT rc;

	ZeroMemory(ret, sizeof(ret));
	ret->lpVtbl = &ShellBrowserVtbl;
	ret->lpIOleCommandTargetVtbl = &ShellBrowser_IOleCommandTargetVtbl;
	ret->refcount = 0;
	ret->hwndSelf = hwnd;

	ret->hwndStatus = CreateWindowW(STATUSCLASSNAMEW, NULL, SBARS_SIZEGRIP|WS_VISIBLE|WS_CHILD, 0, 0, 0, 0, hwnd, NULL, NULL, 0);
	// TODO: remove debug
	old_status_proc = (WNDPROC)GetWindowLongPtrW(ret->hwndStatus, GWLP_WNDPROC);
	SetWindowLongPtrW(ret->hwndStatus, GWLP_WNDPROC, (LONG_PTR)status_proc);

	if (FAILED(SHGetDesktopFolder(&ret->folder)))
		goto failed;
	if (FAILED(IShellFolder_CreateViewObject(ret->folder, ret->hwndSelf, &IID_IShellView, &ret->view)))
		goto failed;

	GetClientRect(hwnd, &rc);
	settings.ViewMode = FVM_ICON;
	settings.fFlags = 0;
	if (FAILED(IShellView_CreateViewWindow(ret->view, NULL, &settings, (IShellBrowser *)ret, &rc, &ret->hwndView)))
		goto failed;

	IShellView_UIActivate(ret->view, SVUIA_ACTIVATE_FOCUS);
	
	return ret;
failed:
	// TODO: release COM objs
	ERR("Creating view failed\n");
	CoTaskMemFree(ret);
	return NULL;
}

HRESULT WINAPI ShellBrowser_QueryInterface(IShellBrowser *iface, REFIID iid, void **ppvObject)
{
	*ppvObject = NULL;

	TRACE("%s\n", debugstr_guid(iid));
	if (IsEqualIID(iid, &IID_IUnknown) || IsEqualIID(iid, &IID_IOleWindow) || IsEqualIID(iid, &IID_IShellBrowser))
	{
		*ppvObject = iface;
	} else if (IsEqualIID(iid, &IID_IOleCommandTarget))
	{
		*ppvObject = _IOleCommandTarget_((ShellBrowser *)iface);
	}

	if (*ppvObject)
	{
		IUnknown_AddRef(iface);
		return S_OK;
	}

	return E_NOINTERFACE;
}

ULONG WINAPI ShellBrowser_AddRef(IShellBrowser *iface)
{
	ShellBrowser *This = (ShellBrowser *)iface;

	TRACE("\n");
	return InterlockedIncrement(&This->refcount);
}

ULONG WINAPI ShellBrowser_Release(IShellBrowser *iface)
{
	ShellBrowser *This = (ShellBrowser *)iface;
	ULONG ret;

	TRACE("\n");
	if ((ret = InterlockedDecrement(&This->refcount)) == 0)
	{
		/* TODO: destructor */
	}
	return ret;
}

HRESULT WINAPI ShellBrowser_GetWindow(IShellBrowser *iface, HWND *phwnd)
{
	ShellBrowser *This = (ShellBrowser *)iface;
	TRACE("(%p): return %p\n", This, This->hwndSelf);
	*phwnd = This->hwndSelf;
	return S_OK;
}

HRESULT WINAPI ShellBrowser_ContextSensitiveHelp(IShellBrowser *iface, BOOL fEnterMode)
{
	FIXME("\n");
	return E_UNEXPECTED;
}

HRESULT WINAPI ShellBrowser_InsertMenusSB(IShellBrowser *iface, HMENU hmenuShared, LPOLEMENUGROUPWIDTHS lpMenuWidths)
{
	HMENU hBrowserMenu;

	TRACE("(%p)->(%p, {%d, %d, %d, %d, %d, %d})\n", iface, hmenuShared, lpMenuWidths->width[0], lpMenuWidths->width[1],
		lpMenuWidths->width[2], lpMenuWidths->width[3], lpMenuWidths->width[4], lpMenuWidths->width[5]);
	hBrowserMenu = LoadMenu(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MENU_BROWSER));
	Shell_MergeMenus(hmenuShared, hBrowserMenu, 0, 0, 0xffff, MM_SUBMENUSHAVEIDS);
	return S_OK;
}

HRESULT WINAPI ShellBrowser_SetMenuSB(IShellBrowser *iface, HMENU hmenuShared, HOLEMENU holemenuRes, HWND hwndActiveObject)
{
	ShellBrowser *This = (ShellBrowser *)iface;

	TRACE("(%p)->(%p, %p)\n", iface, hmenuShared, hwndActiveObject);
	SetMenu(This->hwndSelf, hmenuShared);
	return S_OK;
}

HRESULT WINAPI ShellBrowser_RemoveMenusSB(IShellBrowser *iface, HMENU hmenuShared)
{
	TRACE("(%p)\n", hmenuShared);
	return S_OK;
}

HRESULT WINAPI ShellBrowser_SetStatusTextSB(IShellBrowser *iface, LPCOLESTR pszStatusText)
{
	TRACE("(%p, %ws)\n", pszStatusText);
	return E_NOTIMPL;
}

HRESULT WINAPI ShellBrowser_EnableModelessSB(IShellBrowser *iface, BOOL fEnable)
{
	TRACE("\n");
	return E_NOTIMPL;
}

HRESULT WINAPI ShellBrowser_TranslateAcceleratorSB(IShellBrowser *iface, MSG *pmsg, WORD wID)
{
	TRACE("\n");
	return E_NOTIMPL;
}

HRESULT WINAPI ShellBrowser_BrowseObject(IShellBrowser *iface, LPCITEMIDLIST pidl, UINT wFlags)
{
	TRACE("\n");
	return E_NOTIMPL;
}

HRESULT WINAPI ShellBrowser_GetViewStateStream(IShellBrowser *iface, DWORD grfMode, IStream **ppStrm)
{
	TRACE("\n");
	*ppStrm = NULL;
	return E_NOTIMPL;
}

HRESULT WINAPI ShellBrowser_GetControlWindow(IShellBrowser *iface, UINT id, HWND *phwnd)
{
	ShellBrowser *This = (ShellBrowser *)iface;

	TRACE("(%p)->(%d, %p)\n", iface, id, phwnd);
	*phwnd = NULL;

	switch (id)
	{
	case FCW_STATUS:
		*phwnd = This->hwndStatus;
		break;
	default:
		FIXME("Control not avaibable: %d\n", id);
		return E_INVALIDARG;
	}

	return S_OK;
}

HRESULT WINAPI ShellBrowser_SendControlMsg(IShellBrowser *iface, UINT id, UINT uMsg, WPARAM wParam, LPARAM lParam, LRESULT *pret)
{
	ShellBrowser *This = (ShellBrowser *)iface;
	HWND hwndCtrl;
	LRESULT ret;
	HRESULT hr;

	TRACE("(%p)->(%d, 0x%x, %p, %p)\n", iface, id, uMsg, wParam, lParam);

	hr = ShellBrowser_GetControlWindow(iface, id, &hwndCtrl);
	if (FAILED(hr))
		return hr;

	ret = SendMessageW(hwndCtrl, uMsg, wParam, lParam);
	if (pret)
		*pret = ret;
	return S_OK;
}

HRESULT WINAPI ShellBrowser_QueryActiveShellView(IShellBrowser *iface, IShellView **ppshv)
{
	ShellBrowser *This = (ShellBrowser *)iface;

	TRACE("(%p)->(%p)\n", iface, ppshv);
	*ppshv = This->activeView;
	return E_NOTIMPL;
}

HRESULT WINAPI ShellBrowser_OnViewWindowActive(IShellBrowser *iface, IShellView *pshv)
{
	ShellBrowser *This = (ShellBrowser *)iface;

	TRACE("(%p)->(%p)\n", iface, pshv);
	if (pshv)
		IUnknown_AddRef(pshv);
	This->activeView = pshv;
	return S_OK;
}

HRESULT WINAPI ShellBrowser_SetToolbarItems(IShellBrowser *iface, LPTBBUTTONSB lpButtons, UINT nButtons, UINT uFlags)
{
	FIXME("\n");
	return E_NOTIMPL;
}

/* IOleCommandTarget method */
HRESULT WINAPI ShellBrowser_QueryStatus(IOleCommandTarget *iface, const GUID *pguidCmdGroup, ULONG cCmds, OLECMD prgCmds[], OLECMDTEXT *pCmdText)
{
	FIXME("(%p)->(%s, %d)\n", debugstr_guid(pguidCmdGroup));
	return E_FAIL;
}

/* IOleCommandTarget method */
HRESULT WINAPI ShellBrowser_Exec(IOleCommandTarget *iface, const GUID *pguidCmdGroup,
								 DWORD nCmdID, DWORD nCmdExecOpt, VARIANTARG *pvaIn, VARIANTARG *pvaOut)
{
	FIXME("(%p)->(%s, %d)\n", iface, debugstr_guid(pguidCmdGroup), nCmdID);

	if (pguidCmdGroup && IsEqualIID(pguidCmdGroup, &CGID_Explorer))
	{
		/* Commands sent by the view:
		 * 39 (OLECMDID_PREREFRESH) - 2 in pvaIn
		 * 60 (OLECMDID_ADDTRAVELENTRY) - hwnd in pvaIn (VT_INT_PTR). The SHELL_DefView hwnd is sent before UIActivate and the SysListView32 handle after activate
		 */
		if (nCmdID == OLECMDID_PREREFRESH)
			return S_OK;
		return OLECMDERR_E_NOTSUPPORTED;
	}

	if (pguidCmdGroup && IsEqualIID(pguidCmdGroup, &CGID_Explorer))
	{
		/* Command 77 sent after a resize
		 */
		return OLECMDERR_E_NOTSUPPORTED;
	}

	return OLECMDERR_E_UNKNOWNGROUP;
}

static IShellBrowserVtbl ShellBrowserVtbl =
{
	/* IUnknown */
	ShellBrowser_QueryInterface,
	ShellBrowser_AddRef,
	ShellBrowser_Release,
	/* IOleWindow */
	ShellBrowser_GetWindow,
	ShellBrowser_ContextSensitiveHelp,
	/* IShellBrowser */
	ShellBrowser_InsertMenusSB,
	ShellBrowser_SetMenuSB,
	ShellBrowser_RemoveMenusSB,
	ShellBrowser_SetStatusTextSB,
	ShellBrowser_EnableModelessSB,
	ShellBrowser_TranslateAcceleratorSB,
	ShellBrowser_BrowseObject,
	ShellBrowser_GetViewStateStream,
	ShellBrowser_GetControlWindow,
	ShellBrowser_SendControlMsg,
	ShellBrowser_QueryActiveShellView,
	ShellBrowser_OnViewWindowActive,
	ShellBrowser_SetToolbarItems
};

static HRESULT WINAPI ShellBrowser_IOleCommandTarget_QueryInterface(IOleCommandTarget *iface, REFIID iid, void **ppvObject)
{
	ShellBrowser *This = impl_from_IOleCommandTarget(iface);
	return ShellBrowser_QueryInterface((IShellBrowser *)This, iid, ppvObject);
}

static ULONG WINAPI ShellBrowser_IOleCommandTarget_AddRef(IOleCommandTarget *iface)
{
	ShellBrowser *This = impl_from_IOleCommandTarget(iface);
	return ShellBrowser_AddRef((IShellBrowser *)This);
}

static ULONG WINAPI ShellBrowser_IOleCommandTarget_Release(IOleCommandTarget *iface)
{
	ShellBrowser *This = impl_from_IOleCommandTarget(iface);
	return ShellBrowser_Release((IShellBrowser *)This);
}


static IOleCommandTargetVtbl ShellBrowser_IOleCommandTargetVtbl =
{
	/* IUnknown */
	ShellBrowser_IOleCommandTarget_QueryInterface,
	ShellBrowser_IOleCommandTarget_AddRef,
	ShellBrowser_IOleCommandTarget_Release,
	/* IOleCommandTarget */
	ShellBrowser_QueryStatus,
	ShellBrowser_Exec
};

static LRESULT BROWSER_Size(ShellBrowser *This, WPARAM wParam, LPARAM lParam)
{
	SendMessage(This->hwndStatus, WM_SIZE, 0, 0);

	if (This->hwndView)
	{
		RECT rc, rcStatus;
		GetClientRect(This->hwndStatus, &rcStatus);
		GetClientRect(This->hwndSelf, &rc);
		rc.bottom = max(rc.bottom - rcStatus.bottom, rc.top);
		MoveWindow(This->hwndView, 0, 0, rc.right, rc.bottom, FALSE);
	}
	return 0;
}

static LRESULT WINAPI BROWSER_WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	ShellBrowser *brs = (ShellBrowser *)GetWindowLongPtrW(hWnd, 0);

	if (!brs && msg != WM_CREATE)
		return DefWindowProc(hWnd, msg, wParam, lParam);

	switch (msg)
	{
	case WM_CREATE:
		brs = ShellBrowser_Create(hWnd);
		SetWindowLongPtrW(hWnd, 0, (LONG_PTR)brs);
		break;

	case WM_SIZE:
		return BROWSER_Size(brs, wParam, lParam);

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

HRESULT BROWSER_RegisterClass()
{
    WNDCLASSEX  wndcls;

	ZeroMemory(&wndcls, sizeof(wndcls));
    wndcls.cbSize         = sizeof(wndcls);
    wndcls.lpfnWndProc    = BROWSER_WndProc;
    wndcls.hInstance      = GetModuleHandle(NULL);
    wndcls.hIcon          = LoadIcon(hShell32, MAKEINTRESOURCE(5));
    wndcls.hCursor        = LoadCursor(0, IDC_ARROW);
    wndcls.hbrBackground  = (HBRUSH)(COLOR_WINDOW + 1);
    wndcls.lpszMenuName   = NULL;
	wndcls.lpszClassName  = BROWSEWND_CLASS_NAME;
	wndcls.cbWndExtra     = sizeof(void *);

    if (!RegisterClassEx(&wndcls))
		return E_FAIL;
	return S_OK;
}

HWND BROWSER_CreateWindow()
{
    return CreateWindow(BROWSEWND_CLASS_NAME, L"Test", WS_OVERLAPPEDWINDOW, 100, 100,
                 400, 300, NULL, NULL, GetModuleHandle(NULL), NULL);
}