#define CINTERFACE
#include <windows.h>
#include <shlobj.h>
#include "debug.h"
#include "recyclebin.h"

HINSTANCE hMyHandle;
LONG objCount=0;
CRITICAL_SECTION sectionDebug;

FILE *__debug_out;
char guidbuf[100][100];
int guidbuf_pos=0;
const char *__debug_function_name;
const char *__debug_class;

#define CHECK_IID(name) \
	if (IsEqualIID(guid, &IID_##name)) \
	{ \
	sprintf(guidbuf[pos], #name " [{%08x-%04x-%04x-%02x%02x%02x%02x%02x%02x%02x%02x}]", \
			guid->Data1, guid->Data2, guid->Data3, guid->Data4[0], guid->Data4[1], \
			guid->Data4[2], guid->Data4[3], guid->Data4[4], guid->Data4[5], \
			guid->Data4[6], guid->Data4[7]); \
	}

char *debugstr_guid(REFGUID guid)
{
	int pos=guidbuf_pos;
	guidbuf_pos = (guidbuf_pos+1)%100;
	CHECK_IID(IShellFolder)
	else CHECK_IID(IShellFolder2)
	else CHECK_IID(IShellView)
	else CHECK_IID(IShellIcon)
	else CHECK_IID(IShellIconOverlay)
	else CHECK_IID(IShellDetails)
	else CHECK_IID(IFolderFilter)
	else CHECK_IID(IInternetSecurityManager)
	else CHECK_IID(IPersist)
	else CHECK_IID(IPersistFolder)
	else CHECK_IID(IPersistFolder2)
	else CHECK_IID(IPersistFolder3)
	else CHECK_IID(IPersistIDList)
	else CHECK_IID(IObjectWithSite)
	else CHECK_IID(IZoneIdentifier)
	else CHECK_IID(IDropTarget)
	else CHECK_IID(IExtractIconA)
	else CHECK_IID(IExtractIconW)
	else CHECK_IID(IBrowserFrameOptions)
	else CHECK_IID(IDataObject)
	else CHECK_IID(IContextMenu)
	else CHECK_IID(IQueryInfo)
	else
	sprintf(guidbuf[pos], "{%08x-%04x-%04x-%02x%02x%02x%02x%02x%02x%02x%02x}",
		guid->Data1, guid->Data2, guid->Data3, guid->Data4[0], guid->Data4[1],
		guid->Data4[2], guid->Data4[3], guid->Data4[4], guid->Data4[5],
		guid->Data4[6], guid->Data4[7]);
	return guidbuf[pos];
}

void do_print_trace(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	EnterCriticalSection(&sectionDebug);
	fprintf(__debug_out, "%s:%s ", __debug_class, __debug_function_name);
	fflush(__debug_out);
	vfprintf(__debug_out, fmt, ap);
	fflush(__debug_out);
	LeaveCriticalSection(&sectionDebug);
	va_end(ap);
}

void initDebug()
{
	char filename[200];
	wsprintfA(filename, "F:\\folder-out-%08x-%p.txt", GetCurrentProcessId(), hMyHandle);
	__debug_out = fopen(filename, "w");
	InitializeCriticalSection(&sectionDebug);
}

STDAPI DllCanUnloadNow(void)
{
	TRACE("count=%d\n", objCount);
	if (objCount>0)
		return S_FALSE;
	return S_OK;
}

typedef struct
{
	IClassFactoryVtbl *lpVtbl;
	LONG refCount;
} MyClassFactory;

STDAPI MyClassFactory_QueryInterface(IClassFactory *This, REFIID riid,
            void **ppvObject)
{
	*ppvObject = NULL;
	if (!IsEqualGUID(riid, &IID_IUnknown) && !IsEqualGUID(riid, &IID_IClassFactory))
		return E_NOINTERFACE;
	This->lpVtbl->AddRef(This);
	*ppvObject = This;
	return S_OK;
}

LONG STDMETHODCALLTYPE MyClassFactory_AddRef(IClassFactory *This)
{
	MyClassFactory *obj = (MyClassFactory *)This;
	return InterlockedIncrement(&obj->refCount);
}

LONG STDMETHODCALLTYPE MyClassFactory_Release(IClassFactory *This)
{
	MyClassFactory *obj = (MyClassFactory *)This;
	return InterlockedDecrement(&obj->refCount);
}

STDAPI MyClassFactory_CreateInstance(IClassFactory *This, IUnknown *pUnkOuter, REFIID riid, void **ppvObject)
{
	RecycleBin *recycleBin;
	HRESULT ret;
	TRACE("(%p, %s, %p)\n", pUnkOuter, debugstr_guid(riid), ppvObject);

	if (pUnkOuter != NULL)
		return CLASS_E_NOAGGREGATION;
	ret = RecycleBin_Constructor(&recycleBin);
	if (FAILED(ret)) return ret;
	ret = ((IUnknown*)recycleBin)->lpVtbl->QueryInterface((IUnknown*)recycleBin, riid, ppvObject);
	if (FAILED(ret))
	{
		RecycleBin_Destructor(recycleBin);
		ppvObject = NULL;
	}
	TRACE("ret=%08x\n", ret);
	return ret;
}

STDAPI MyClassFactory_LockServer(IClassFactory *This, BOOL fLock)
{
	if (fLock)
		InterlockedIncrement(&objCount);
	else
		InterlockedDecrement(&objCount);
	return S_OK;
}

IClassFactoryVtbl myClassFactoryVtbl =
{
	MyClassFactory_QueryInterface,
	MyClassFactory_AddRef,
	MyClassFactory_Release,
	MyClassFactory_CreateInstance,
	MyClassFactory_LockServer
};

MyClassFactory myClassFactory = { &myClassFactoryVtbl, 0 };

STDAPI DllGetClassObject(REFCLSID clsid, REFIID iid, LPVOID *ppv)
{
	*ppv=NULL;
	TRACE("(%s, %s, %p)\n", debugstr_guid(clsid), debugstr_guid(iid), ppv);
	if (!IsEqualGUID(iid, &IID_IClassFactory))
		return E_INVALIDARG;

	*ppv=&myClassFactory;
	return S_OK;
}

BOOL WINAPI DllMain(HINSTANCE hInstDll, DWORD fdwReason, LPVOID reserved)
{
	switch (fdwReason)
	{
	case DLL_PROCESS_ATTACH:
		hMyHandle = hInstDll;
		initDebug();
		TRACE("ProcessAttach\n");
		DisableThreadLibraryCalls(hInstDll);
		return TRUE;
	case DLL_PROCESS_DETACH:
		TRACE("ProcessDetach\n");
		return TRUE;
	}
	return TRUE;
}