#define CINTERFACE
#include <windows.h>
#include <ntquery.h>
#include <shlwapi.h>
#include <shlobj.h>
#include "recyclebin.h"
#include "debug.h"

#define IUnknown_AddRef(p) (p)->lpVtbl->AddRef(p)
#define IUnknown_QueryInterface(p,a,b) (p)->lpVtbl->QueryInterface(p, a, b)

typedef struct
{
	const char *column_name;
	const GUID *fmtId;
	DWORD pid;
	int pcsFlags;
	int fmt;
	int cxChars;
} columninfo;

static columninfo RecycleBinColumns[] =
{
	{"QnameQ",    &FMTID_Storage,   PID_STG_NAME,       SHCOLSTATE_TYPE_STR|SHCOLSTATE_ONBYDEFAULT,  LVCFMT_LEFT,  30},
	{"QdelfromQ", &FMTID_Displaced, PID_DISPLACED_FROM, SHCOLSTATE_TYPE_STR|SHCOLSTATE_ONBYDEFAULT,  LVCFMT_LEFT,  30},
	{"QdatedelQ", &FMTID_Displaced, PID_DISPLACED_DATE, SHCOLSTATE_TYPE_DATE|SHCOLSTATE_ONBYDEFAULT, LVCFMT_LEFT,  20},
	{"QsizeQ",    &FMTID_Storage,   PID_STG_SIZE,       SHCOLSTATE_TYPE_INT|SHCOLSTATE_ONBYDEFAULT,  LVCFMT_RIGHT, 20},
	{"QtypeQ",    &FMTID_Storage,   PID_STG_STORAGETYPE,SHCOLSTATE_TYPE_INT|SHCOLSTATE_ONBYDEFAULT,  LVCFMT_LEFT,  20},
	{"QmtimeQ",   &FMTID_Storage,   PID_STG_WRITETIME,  SHCOLSTATE_TYPE_DATE|SHCOLSTATE_ONBYDEFAULT, LVCFMT_LEFT,  20},
	{"QcrtimeQ",  &FMTID_Storage,   PID_STG_CREATETIME, SHCOLSTATE_TYPE_DATE,                        LVCFMT_LEFT,  20},
/*	{"QattribQ",  &FMTID_Storage,   PID_STG_ATTRIBUTES, SHCOLSTATE_TYPE_STR,                         LVCFMT_LEFT,  20},*/
};

#define COLUMN_NAME    0
#define COLUMN_DELFROM 1
#define COLUMN_DATEDEL 2
#define COLUMN_SIZE    3
#define COLUMN_TYPE    4
#define COLUMN_MTIME   5
#define COLUMN_CRTIME  6

#define COLUMNS_COUNT  7

DEFINE_GUID(CLSID_RecycyleBin, 0xc0c0c0c0,0xc0c0,0xc0c0,0xc0,0xc0,0xc0,0xc0,0xc0,0xc0,0xc0,0xc0);

#define SHAlloc CoTaskMemAlloc
#define SHFree CoTaskMemFree

HRESULT FormatDateTime(LPWSTR buffer, int size, FILETIME ft)
{
	FILETIME lft;
	SYSTEMTIME time;
	int ret;

	FileTimeToLocalFileTime(&ft, &lft);
    FileTimeToSystemTime(&lft, &time);

    ret = GetDateFormatW(LOCALE_USER_DEFAULT, DATE_SHORTDATE, &time, NULL, buffer, size);
    if (ret>0 && ret<size)
    {
        /* Append space + time without seconds */
        buffer[ret-1] = ' ';
        GetTimeFormatW(LOCALE_USER_DEFAULT, TIME_NOSECONDS, &time, NULL, &buffer[ret], size - ret);
    }

	return (ret!=0 ? E_FAIL : S_OK);
}

/*
 * Enum ID List
 */
typedef struct tagRecycleBinEnumIDList
{
	IEnumIDListVtbl *lpVtbl;
	LONG refCount;
	int pos;
	LPITEMIDLIST *pidls;
	int count;
} RecycleBinEnumIDList;

IEnumIDListVtbl recycleBinEnumIDListVtbl;

HRESULT RecycleBinEnumIDList_Constructor(RecycleBinEnumIDList **ppvOut)
{
	HRESULT ret;
	*ppvOut = SHAlloc(sizeof(RecycleBinEnumIDList));
	if (*ppvOut == NULL) return E_OUTOFMEMORY;
	ZeroMemory(*ppvOut, sizeof(RecycleBinEnumIDList));
	(*ppvOut)->lpVtbl = &recycleBinEnumIDListVtbl;
	(*ppvOut)->refCount = 1;
	if (FAILED(ret = TRASH_EnumElements(&(*ppvOut)->pidls, &(*ppvOut)->count)))
	{
		SHFree(*ppvOut);
		*ppvOut = NULL;
		return ret;
	}
	InterlockedIncrement(&objCount);
	return S_OK;
}

void RecycleBinEnumIDList_Destructor(RecycleBinEnumIDList *This)
{
	int i;
	for (i=0; i<This->count; i++)
		SHFree(This->pidls[i]);
	SHFree(This->pidls);
	SHFree(This);
	InterlockedDecrement(&objCount);
}

HRESULT RecycleBinEnumIDList_QueryInterface(IEnumIDList *iface, REFIID riid, void **ppvObject)
{
	RecycleBinEnumIDList *This = (RecycleBinEnumIDList *)iface;
	TRACE("(%p, %s, %p)\n", This, debugstr_guid(riid), ppvObject);

	*ppvObject = NULL;
	if (IsEqualGUID(riid, &IID_IUnknown) || IsEqualGUID(riid, &IID_IEnumIDList))
	{
		*ppvObject = This;
		IUnknown_AddRef((IUnknown *)This);
		return S_OK;
	}

	WARN("no interface\n");
	return E_NOINTERFACE;
}

LONG RecycleBinEnumIDList_AddRef(IEnumIDList *iface)
{
	RecycleBinEnumIDList *This = (RecycleBinEnumIDList *)iface;
	TRACE("(%p)\n", This);
	return InterlockedIncrement(&This->refCount);
}

LONG RecycleBinEnumIDList_Release(IEnumIDList *iface)
{
	RecycleBinEnumIDList *This = (RecycleBinEnumIDList *)iface;
	LONG result;

	TRACE("(%p)\n", This);
	result = InterlockedDecrement(&This->refCount);
	if (result == 0)
	{
		TRACE("Destroy object\n");
		RecycleBinEnumIDList_Destructor(This);
	}
	return result;
}

HRESULT RecycleBinEnumIDList_Next(IEnumIDList *iface, ULONG celt, LPITEMIDLIST *rgelt, ULONG *pceltFetched)
{
	RecycleBinEnumIDList *This = (RecycleBinEnumIDList *)iface;
	ULONG i;

	TRACE("(%p, %u, %p, %p)\n", iface, celt, rgelt, pceltFetched);
	*pceltFetched = 0;

	ZeroMemory(rgelt, sizeof(LPITEMIDLIST)*celt);
	for (i=0; i<celt; i++)
	{
		int num = This->pos++;
		if (num >= This->count)
			break;
		rgelt[i] = ILClone(This->pidls[num]);
		TRACE("-> %p\n", rgelt[i]);
	}
	if (pceltFetched)
		*pceltFetched = i;
	return (i==celt ? S_OK : S_FALSE);
}

HRESULT RecycleBinEnumIDList_Skip(IEnumIDList *iface, ULONG celt)
{
	RecycleBinEnumIDList *This = (RecycleBinEnumIDList *)iface;
	TRACE("(%p)\n", iface);

	This->pos += celt;
	return S_OK;     /* FIXME: test */
}

HRESULT RecycleBinEnumIDList_Reset(IEnumIDList *iface)
{
	RecycleBinEnumIDList *This = (RecycleBinEnumIDList *)iface;
	TRACE("(%p)\n", iface);

	This->pos = 0;
	return S_OK;
}

HRESULT RecycleBinEnumIDList_Clone(IEnumIDList *This, IEnumIDList **ppenum)
{
	TRACE("\n");
	*ppenum = NULL;
	return E_NOTIMPL;
}

IEnumIDListVtbl recycleBinEnumIDListVtbl =
{
	/* IUnknown */
	RecycleBinEnumIDList_QueryInterface,
	RecycleBinEnumIDList_AddRef,
	RecycleBinEnumIDList_Release,

	/* IEnumIDList */
	RecycleBinEnumIDList_Next,
	RecycleBinEnumIDList_Skip,
	RecycleBinEnumIDList_Reset,
	RecycleBinEnumIDList_Clone
};

/*
 * Recycle Bin folder
 */

struct tagRecycleBin
{
	IShellFolder2Vtbl *lpVtbl;
	IPersistFolder2Vtbl *lpPersistFolderVtbl;
	IShellFolderViewCBVtbl *lpCallbackVtbl;
	IDropTargetVtbl *lpDropTargetVtbl;
	LONG refCount;

	LPITEMIDLIST pidl;
};

extern IShellFolder2Vtbl recycleBinVtbl;
extern IPersistFolder2Vtbl recycleBinPersistVtbl;
extern IShellFolderViewCBVtbl recycleBinCallbackVtbl;
extern IDropTargetVtbl recycleBinDropTargetVtbl;

static RecycleBin *impl_from_IPersistFolder(IPersistFolder2 *iface)
{
    return (RecycleBin *)((char*)iface - FIELD_OFFSET(RecycleBin, lpPersistFolderVtbl));
}

static RecycleBin *impl_from_IShellFolderViewCB(IShellFolderViewCB *iface)
{
    return (RecycleBin *)((char*)iface - FIELD_OFFSET(RecycleBin, lpCallbackVtbl));
}

static RecycleBin *impl_from_IDropTarget(IDropTarget *iface)
{
    return (RecycleBin *)((char*)iface - FIELD_OFFSET(RecycleBin, lpDropTargetVtbl));
}

HRESULT RecycleBin_Constructor(RecycleBin **ppOutput)
{
	RecycleBin *ret = SHAlloc(sizeof(RecycleBin));
	if (ret == NULL)
		return E_OUTOFMEMORY;
	ZeroMemory(ret, sizeof(RecycleBin));
	ret->lpVtbl = &recycleBinVtbl;
	ret->lpPersistFolderVtbl = &recycleBinPersistVtbl;
	ret->lpCallbackVtbl = &recycleBinCallbackVtbl;
	ret->lpDropTargetVtbl = &recycleBinDropTargetVtbl;
	*ppOutput = ret;
	InterlockedIncrement(&objCount);
	return S_OK;
}

void RecycleBin_Destructor(RecycleBin *This)
{
	InterlockedDecrement(&objCount);
	SHFree(This->pidl);
	SHFree(This);
}

HRESULT RecycleBin_QueryInterface(IShellFolder2 *iface, REFIID riid, void **ppvObject)
{
	RecycleBin *This = (RecycleBin *)iface;
	TRACE("(%p, %s, %p)\n", This, debugstr_guid(riid), ppvObject);

	*ppvObject = NULL;
	if (IsEqualGUID(riid, &IID_IUnknown) || IsEqualGUID(riid, &IID_IShellFolder)
			|| IsEqualGUID(riid, &IID_IShellFolder2))
		*ppvObject = This;

	if (IsEqualGUID(riid, &IID_IPersist) || IsEqualGUID(riid, &IID_IPersistFolder) ||
			IsEqualGUID(riid, &IID_IPersistFolder2))
		*ppvObject = &This->lpPersistFolderVtbl;

	if (IsEqualGUID(riid, &IID_IDropTarget))
		*ppvObject = &This->lpDropTargetVtbl;

	if (*ppvObject != NULL)
	{
		IUnknown_AddRef((IUnknown *)*ppvObject);
		return S_OK;
	}
	WARN("no interface\n");
	return E_NOINTERFACE;
}

LONG RecycleBin_AddRef(IShellFolder2 *iface)
{
	RecycleBin *This = (RecycleBin *)iface;
	TRACE("(%p)\n", This);
	return InterlockedIncrement(&This->refCount);
}

LONG RecycleBin_Release(IShellFolder2 *iface)
{
	RecycleBin *This = (RecycleBin *)iface;
	LONG result;

	TRACE("(%p)\n", This);
	result = InterlockedDecrement(&This->refCount);
	if (result == 0)
	{
		TRACE("Destroy object\n");
		RecycleBin_Destructor(This);
	}
	return result;
}

HRESULT RecycleBin_ParseDisplayName(IShellFolder2 *This, HWND hwnd, LPBC pbc,
            LPOLESTR pszDisplayName, ULONG *pchEaten, LPITEMIDLIST *ppidl,
            ULONG *pdwAttributes)
{
	FIXME("stub\n");
	return E_NOTIMPL;
}

HRESULT RecycleBin_EnumObjects(IShellFolder2 *iface, HWND hwnd, SHCONTF grfFlags, IEnumIDList **ppenumIDList)
{
	RecycleBin *This = (RecycleBin *)iface;
	TRACE("(%p, %p, %x, %p)\n", This, hwnd, grfFlags, ppenumIDList);
	return RecycleBinEnumIDList_Constructor((RecycleBinEnumIDList **)ppenumIDList);
}

HRESULT RecycleBin_BindToObject(IShellFolder2 *This, LPCITEMIDLIST pidl, LPBC pbc, REFIID riid, void **ppv)
{
	FIXME("(%p, %p, %p, %s, %p) - stub\n", This, pidl, pbc, debugstr_guid(riid), ppv);
	return E_NOTIMPL;
}

HRESULT RecycleBin_BindToStorage(IShellFolder2 *This, LPCITEMIDLIST pidl, LPBC pbc, REFIID riid, void **ppv)
{
	FIXME("(%p, %p, %p, %s, %p) - stub\n", This, pidl, pbc, debugstr_guid(riid), ppv);
	return E_NOTIMPL;
}

HRESULT RecycleBin_CompareIDs(IShellFolder2 *iface, LPARAM lParam, LPCITEMIDLIST pidl1, LPCITEMIDLIST pidl2)
{
	RecycleBin *This = (RecycleBin *)iface;

	/* TODO */
	TRACE("(%p, %x, %p, %p)\n", iface, lParam, pidl1, pidl2);
	if (pidl1->mkid.cb != pidl2->mkid.cb)
		return MAKE_HRESULT(SEVERITY_SUCCESS, 0, pidl1->mkid.cb - pidl2->mkid.cb);
	return MAKE_HRESULT(SEVERITY_SUCCESS, 0, (unsigned short)memcmp(pidl1->mkid.abID, pidl2->mkid.abID, pidl1->mkid.cb));
}

HRESULT RecycleBin_CreateViewObject(IShellFolder2 *iface, HWND hwndOwner, REFIID riid, void **ppv)
{
	RecycleBin *This = (RecycleBin *)iface;
	HRESULT ret;
	TRACE("(%p, %p, %s, %p)\n", This, hwndOwner, debugstr_guid(riid), ppv);

	*ppv = NULL;
	if (IsEqualGUID(riid, &IID_IShellView))
	{
		IShellView *tmp;
		SFV_CREATE sfv;

		ZeroMemory(&sfv, sizeof(sfv));
		sfv.cbSize = sizeof(sfv);
		sfv.pshf = (IShellFolder *)This;
		sfv.psfvcb = (IShellFolderViewCB *)&This->lpCallbackVtbl;

		TRACE("Calling SHCreateShellFolderView\n");
		ret = SHCreateShellFolderView(&sfv, &tmp);
		TRACE("Result: %08x, output: %p\n", ret, tmp);
		*ppv = tmp;
		return ret;
	}

	if (IsEqualGUID(riid, &IID_IDropTarget))
	{
		return IUnknown_QueryInterface(iface, riid, ppv);
	}
	return E_NOINTERFACE;
}

HRESULT RecycleBin_GetAttributesOf(IShellFolder2 *This, UINT cidl, LPCITEMIDLIST *apidl,
								   SFGAOF *rgfInOut)
{
	TRACE("(%p, %d, {%p, ...}, {%x})\n", This, cidl, apidl[0], *rgfInOut);
	*rgfInOut &= SFGAO_CANMOVE|SFGAO_CANDELETE|SFGAO_HASPROPSHEET|SFGAO_FILESYSTEM;
	return S_OK;
}

HRESULT RecycleBin_GetUIObjectOf(IShellFolder2 *This, HWND hwndOwner, UINT cidl, LPCITEMIDLIST *apidl,
					  REFIID riid, UINT *rgfReserved, void **ppv)
{
	FIXME("(%p, %p, %d, {%p, ...}, %s, %p, %p): stub!\n", This, hwndOwner, cidl, apidl[0], debugstr_guid(riid), rgfReserved, ppv);
	*ppv = NULL;
	return E_NOTIMPL;
}

HRESULT RecycleBin_GetDisplayNameOf(IShellFolder2 *This, LPCITEMIDLIST pidl, SHGDNF uFlags, STRRET *pName)
{
	int len = pidl->mkid.cb-1;
	WIN32_FIND_DATAW data;
	WCHAR buf[256];

	TRACE("(%p, %p, %x, %p)\n", This, pidl, uFlags, pName);
	TRASH_UnpackItemID(&pidl->mkid, NULL, &data);
	pName->uType = STRRET_WSTR;
	LoadStringW(GetModuleHandleW(L"shellfolder.dll"), 200, buf, 256);
	pName->pOleStr = StrDupW(/*PathFindFileNameW(data.cFileName)*/buf);
	if (pName->pOleStr == NULL)
		return E_OUTOFMEMORY;

	return S_OK;
}

HRESULT RecycleBin_SetNameOf(IShellFolder2 *This, HWND hwnd, LPCITEMIDLIST pidl, LPCOLESTR pszName,
            SHGDNF uFlags, LPITEMIDLIST *ppidlOut)
{
	TRACE("\n");
	return E_FAIL; /* not supported */
}

HRESULT RecycleBin_GetClassID(IPersistFolder2 *This, CLSID *pClassID)
{
	TRACE("(%p, %p)\n", This, pClassID);
	if (This == NULL || pClassID == NULL)
		return E_INVALIDARG;
	memcpy(pClassID, &CLSID_RecycleBin, sizeof(CLSID));
	return S_OK;
}

HRESULT RecycleBin_Initialize(IPersistFolder2 *iface, LPCITEMIDLIST pidl)
{
	RecycleBin *This = impl_from_IPersistFolder(iface);
	TRACE("(%p, %p)\n", pidl);
	
	This->pidl = ILClone(pidl);
	if (This->pidl == NULL)
		return E_OUTOFMEMORY;
	return S_OK;
}

HRESULT RecycleBin_GetCurFolder(IPersistFolder2 *iface, LPITEMIDLIST *ppidl)
{
	RecycleBin *This = impl_from_IPersistFolder(iface);
	TRACE("\n");
	*ppidl = ILClone(This->pidl);
	return S_OK;
}

HRESULT RecycleBin_GetDefaultSearchGUID(IShellFolder2 *iface, GUID *pguid)
{
	TRACE("\n");
	return E_NOTIMPL;
}

HRESULT RecycleBin_EnumSearches(IShellFolder2 *iface, IEnumExtraSearch **ppEnum)
{
	TRACE("\n");
	*ppEnum = NULL;
	return E_NOTIMPL;
}

HRESULT RecycleBin_GetDefaultColumn(IShellFolder2 *iface, DWORD dwReserved, ULONG *pSort, ULONG *pDisplay)
{
	RecycleBin *This = (RecycleBin *)iface;
	TRACE("(%p, %x, %p, %p)\n", This, dwReserved, pSort, pDisplay);
	*pSort = 0;
	*pDisplay = 0;
	return S_OK;
}

HRESULT RecycleBin_GetDefaultColumnState(IShellFolder2 *iface, UINT iColumn, SHCOLSTATEF *pcsFlags)
{
	RecycleBin *This = (RecycleBin *)iface;
	TRACE("(%p, %d, %p, %p)\n", This, iColumn, pcsFlags);
	if (iColumn < 0 || iColumn >= COLUMNS_COUNT)
		return E_INVALIDARG;
	*pcsFlags = RecycleBinColumns[iColumn].pcsFlags;
	return S_OK;
}

HRESULT RecycleBin_GetDetailsEx(IShellFolder2 *iface, LPCITEMIDLIST pidl, const SHCOLUMNID *pscid, VARIANT *pv)
{
	TRACE("\n");
	return E_NOTIMPL;
}

HRESULT RecycleBin_GetDetailsOf(IShellFolder2 *iface, LPCITEMIDLIST pidl, UINT iColumn, LPSHELLDETAILS pDetails)
{
	RecycleBin *This = (RecycleBin *)iface;
	WIN32_FIND_DATA data;
	WCHAR buffer[MAX_PATH];

	TRACE("(%p, %p, %d, %p)\n", This, pidl, iColumn, pDetails);
	if (iColumn < 0 || iColumn >= COLUMNS_COUNT)
		return E_FAIL;
	pDetails->fmt = RecycleBinColumns[iColumn].fmt;
	pDetails->cxChar = RecycleBinColumns[iColumn].cxChars;
	if (pidl == NULL)
	{
		pDetails->str.uType = STRRET_CSTR;
		lstrcpyA(pDetails->str.cStr, RecycleBinColumns[iColumn].column_name);
		return S_OK;
	}

	if (iColumn == COLUMN_NAME)
		return RecycleBin_GetDisplayNameOf(iface, pidl, SHGDN_NORMAL, &pDetails->str);

	TRASH_UnpackItemID(&pidl->mkid, NULL, &data);
	switch (iColumn)
	{
		case COLUMN_DATEDEL:
			FormatDateTime(buffer, MAX_PATH, data.ftLastAccessTime);
			break;
		case COLUMN_DELFROM:
			lstrcpyW(buffer, data.cFileName);
			PathRemoveFileSpec(buffer);
			break;
		case COLUMN_SIZE:
			StrFormatByteSizeW(((LONGLONG)data.nFileSizeHigh<<32)|data.nFileSizeLow, buffer, MAX_PATH);
			break;
		case COLUMN_MTIME:
			FormatDateTime(buffer, MAX_PATH, data.ftLastWriteTime);
			break;
		case COLUMN_CRTIME:
			FormatDateTime(buffer, MAX_PATH, data.ftCreationTime);
			break;
		case COLUMN_TYPE:
			/* TODO */
			buffer[0] = 0;
			break;
		default:
			return E_FAIL;
	}
	
	pDetails->str.uType = STRRET_WSTR;
	pDetails->str.pOleStr = StrDupW(buffer);
	return (pDetails->str.pOleStr != NULL ? S_OK : E_OUTOFMEMORY);
}

HRESULT RecycleBin_MapColumnToSCID(IShellFolder2 *iface, UINT iColumn, SHCOLUMNID *pscid)
{
	RecycleBin *This = (RecycleBin *)iface;
	TRACE("(%p, %d, %p)\n", This, iColumn, pscid);
	if (iColumn<0 || iColumn>=COLUMNS_COUNT)
		return E_INVALIDARG;
	pscid->fmtid = *RecycleBinColumns[iColumn].fmtId;
	pscid->pid = RecycleBinColumns[iColumn].pid;
	return S_OK;
}

HRESULT RecycleBin_DragEnter(IDropTarget *iface, IDataObject *pDataObj, DWORD grfKeyState,
            POINTL pt, DWORD *pdwEffect)
{
	TRACE("\n");
	*pdwEffect = DROPEFFECT_NONE;
	return S_OK;
}

HRESULT RecycleBin_DragOver(IDropTarget *iface, DWORD grfKeyState, POINTL pt,
			DWORD *pdwEffect)
{
	TRACE("\n");
	*pdwEffect = DROPEFFECT_NONE;
	return S_OK;
}

HRESULT RecycleBin_DragLeave(IDropTarget *iface)
{
	TRACE("\n");
	return S_OK;
}

HRESULT RecycleBin_Drop(IDropTarget * This, IDataObject *pDataObj, DWORD grfKeyState,
            POINTL pt, DWORD *pdwEffect)
{
	TRACE("\n");
	return S_OK;
}

HRESULT RecycleBin_MessageSFVCB(IShellFolderViewCB *iface, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	TRACE("(%p, %x, %p, %p)\n", iface, uMsg, wParam, lParam);
	switch (uMsg)
	{
		case SFVM_GETSORTDEFAULTS:
		{
			int *iDirection = (int *)wParam;
			int *iColumn = (int *)lParam;

			TRACE("SFVM_GETSORTDEFAULTS\n");
			if (iDirection)
				*iDirection = 0;
			if (iColumn)
				*iColumn = 0;

			return S_OK;
		}
	}
	return E_NOTIMPL;
}


IShellFolder2Vtbl recycleBinVtbl = 
{
	/* IUnknown */
	RecycleBin_QueryInterface,
	RecycleBin_AddRef,
	RecycleBin_Release,

	/* IShellFolder */
	RecycleBin_ParseDisplayName,
	RecycleBin_EnumObjects,
	RecycleBin_BindToObject,
	RecycleBin_BindToStorage,
	RecycleBin_CompareIDs,
	RecycleBin_CreateViewObject,
	RecycleBin_GetAttributesOf,
	RecycleBin_GetUIObjectOf,
	RecycleBin_GetDisplayNameOf,
	RecycleBin_SetNameOf,

	/* IShellFolder2 */
	RecycleBin_GetDefaultSearchGUID,
	RecycleBin_EnumSearches,
	RecycleBin_GetDefaultColumn,
	RecycleBin_GetDefaultColumnState,
	RecycleBin_GetDetailsEx,
	RecycleBin_GetDetailsOf,
	RecycleBin_MapColumnToSCID
};

HRESULT RecycleBin_IPersistFolder_QueryInterface(IPersistFolder2 *This, REFIID riid, void **ppvObject)
{
	return RecycleBin_QueryInterface((IShellFolder2 *)impl_from_IPersistFolder(This), riid, ppvObject);
}

LONG RecycleBin_IPersistFolder_AddRef(IPersistFolder2 *This)
{
	return RecycleBin_AddRef((IShellFolder2 *)impl_from_IPersistFolder(This));
}

LONG RecycleBin_IPersistFolder_Release(IPersistFolder2 *This)
{
	return RecycleBin_Release((IShellFolder2 *)impl_from_IPersistFolder(This));
}

IPersistFolder2Vtbl recycleBinPersistVtbl =
{
	/* IUnknown */
	RecycleBin_IPersistFolder_QueryInterface,
	RecycleBin_IPersistFolder_AddRef,
	RecycleBin_IPersistFolder_Release,

	/* IPersist */
	RecycleBin_GetClassID,
	/* IPersistFolder */
	RecycleBin_Initialize,
	/* IPersistFolder2 */
	RecycleBin_GetCurFolder
};

HRESULT RecycleBin_IShellFolderViewCB_QueryInterface(IShellFolderViewCB *This, REFIID riid, void **ppvObject)
{
	return RecycleBin_QueryInterface((IShellFolder2 *)impl_from_IShellFolderViewCB(This), riid, ppvObject);
}

LONG RecycleBin_IShellFolderViewCB_AddRef(IShellFolderViewCB *This)
{
	return RecycleBin_AddRef((IShellFolder2 *)impl_from_IShellFolderViewCB(This));
}

LONG RecycleBin_IShellFolderViewCB_Release(IShellFolderViewCB *This)
{
	return RecycleBin_Release((IShellFolder2 *)impl_from_IShellFolderViewCB(This));
}

IShellFolderViewCBVtbl recycleBinCallbackVtbl =
{
	/* IUnknown */
	RecycleBin_IShellFolderViewCB_QueryInterface,
	RecycleBin_IShellFolderViewCB_AddRef,
	RecycleBin_IShellFolderViewCB_Release,

	RecycleBin_MessageSFVCB
};

HRESULT RecycleBin_IDropTarget_QueryInterface(IDropTarget *This, REFIID riid, void **ppvObject)
{
	return RecycleBin_QueryInterface((IShellFolder2 *)impl_from_IDropTarget(This), riid, ppvObject);
}

LONG RecycleBin_IDropTarget_AddRef(IDropTarget *This)
{
	return RecycleBin_AddRef((IShellFolder2 *)impl_from_IDropTarget(This));
}

LONG RecycleBin_IDropTarget_Release(IDropTarget *This)
{
	return RecycleBin_Release((IShellFolder2 *)impl_from_IDropTarget(This));
}

IDropTargetVtbl recycleBinDropTargetVtbl =
{
	/* IUnknown */
	RecycleBin_IDropTarget_QueryInterface,
	RecycleBin_IDropTarget_AddRef,
	RecycleBin_IDropTarget_Release,

	RecycleBin_DragEnter,
	RecycleBin_DragOver,
	RecycleBin_DragLeave,
	RecycleBin_Drop
};
