#include <windows.h>
#include <shlwapi.h>
#include <shlobj.h>
#include "recyclebin.h"
#include "debug.h"

#define SHAlloc CoTaskMemAlloc
#define SHFree CoTaskMemFree

typedef struct tagTRASH_BUCKET TRASH_BUCKET;

TRASH_BUCKET *home_bucket = NULL;

typedef struct tagTRASH_ELEMENT
{
	TRASH_BUCKET *bucket;
	char *filename;
} TRASH_ELEMENT;

HRESULT TRASH_CreateSimplePIDL(const TRASH_ELEMENT *element, const WIN32_FIND_DATAW *data, LPITEMIDLIST *pidlOut)
{
	LPITEMIDLIST pidl = SHAlloc(2+1+sizeof(WIN32_FIND_DATAW)+1+lstrlenA(element->filename)+1+2);
	*pidlOut = NULL;
	if (pidl == NULL)
		return E_OUTOFMEMORY;
	pidl->mkid.cb = (USHORT)(2+1+sizeof(WIN32_FIND_DATAW)+1+lstrlenA(element->filename)+1);
	pidl->mkid.abID[0] = 0;
	memcpy(pidl->mkid.abID+1, data, sizeof(WIN32_FIND_DATAW));
	pidl->mkid.abID[1+sizeof(WIN32_FIND_DATAW)] = 0;
	lstrcpyA(pidl->mkid.abID+1+sizeof(WIN32_FIND_DATAW)+1, element->filename);
	*(USHORT *)(pidl->mkid.abID+1+sizeof(WIN32_FIND_DATAW)+1+lstrlenA(element->filename)+1) = 0;
	*pidlOut = pidl;
	return S_OK;
}

HRESULT TRASH_UnpackItemID(LPCSHITEMID id, TRASH_ELEMENT *element, WIN32_FIND_DATAW *data)
{
	if (id->cb < 2+1+sizeof(WIN32_FIND_DATAW)+2)
		return E_INVALIDARG;
	if (id->abID[0] != 0 || id->abID[1+sizeof(WIN32_FIND_DATAW)] != 0)
		return E_INVALIDARG;
	if (memchr(&id->abID+1+sizeof(WIN32_FIND_DATAW)+1, 0, id->cb-(2+1+sizeof(WIN32_FIND_DATAW)+1)) == NULL)
		return E_INVALIDARG;

	if (data != NULL)
		*data = *(WIN32_FIND_DATAW *)(id->abID+1);
	if (element != NULL)
	{
		element->bucket = home_bucket;
		element->filename = StrDupA(id->abID+1+sizeof(WIN32_FIND_DATAW)+1);
		if (element->filename == NULL)
			return E_OUTOFMEMORY;
	}
	return S_OK;
}

static const int CONST_numElems=5;
static const char *CONST_elemsIds[]={"aaa", "bbb", "c123", "dqq", "ezaz"};

HRESULT TRASH_EnumElements(LPITEMIDLIST **pidls, int *count)
{
	LPITEMIDLIST *out = SHAlloc(CONST_numElems*sizeof(LPITEMIDLIST));
	int i;
	*pidls = NULL;

	if (out == NULL)
		return E_OUTOFMEMORY;
	for (i=0; i<CONST_numElems; i++)
	{
		WIN32_FIND_DATAW data;
		TRASH_ELEMENT elem;

		ZeroMemory(&data, sizeof(data));
		data.nFileSizeLow = i*104857600;
		data.nFileSizeHigh = 2;
		if (i%2)
			wsprintfW(data.cFileName, L"C:\\file_%hs", CONST_elemsIds[i]);
		else
			wsprintfW(data.cFileName, L"D:\\Test\\file_%hs", CONST_elemsIds[i]);
		CoFileTimeNow(&data.ftLastAccessTime);
		CoFileTimeNow(&data.ftLastWriteTime);
		CoFileTimeNow(&data.ftCreationTime);

		elem.bucket = NULL;
		elem.filename = CONST_elemsIds[i];
		if (elem.filename == NULL)
		{
			SHFree(out);
			return E_OUTOFMEMORY;
		}
		TRASH_CreateSimplePIDL(&elem, &data, &out[i]);
	}
	*pidls = out;
	*count = CONST_numElems;
	return S_OK;
}